use std::{net::SocketAddr, process};

use axum::{routing::post, Router};

use sqlx::postgres::PgPoolOptions;
use tokio::signal;

use tower_http::trace::TraceLayer;
use tracing_subscriber::{
    filter::LevelFilter, layer::SubscriberExt, util::SubscriberInitExt, Layer,
};

use god_replacement_product::{
    discord::{commands::register_commands, interactions::post_interaction},
    AppState, Config,
};

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    let config = Config::configure()?;
    let tracer = opentelemetry_jaeger::new_agent_pipeline()
        .with_service_name("god_replacement_product")
        .install_simple()?;
    let telemetry = tracing_opentelemetry::layer().with_tracer(tracer);
    let (loki_layer, loki_task) = tracing_loki::builder()
        .label("host", "mine")?
        .extra_field("pid", format!("{}", process::id()))?
        .build_url(url::Url::parse("http://127.0.0.1:3100").unwrap())?;

    tracing_subscriber::registry()
        .with(telemetry.with_filter(LevelFilter::DEBUG))
        .with(
            tracing_subscriber::fmt::layer()
                .and_then(loki_layer)
                .with_filter(
                    tracing_subscriber::EnvFilter::try_from_default_env().unwrap_or_else(|_| {
                        "god_replacement_product=debug,tower_http=debug".into()
                    }),
                ),
        )
        .init();

    let port = config.listen_port;

    let pg_pool = PgPoolOptions::new()
        .max_connections(5)
        .connect(&config.database_url)
        .await?;

    sqlx::migrate!().run(&pg_pool).await?;

    register_commands(config.discord_client_id.to_owned(), config.authorization()).await?;
    let state = AppState { config, pg_pool };

    let app = Router::new()
        .route("/api/discord/interactions/", post(post_interaction))
        .with_state(state)
        .layer(TraceLayer::new_for_http());

    let addr = SocketAddr::from(([127, 0, 0, 1], port));
    tracing::debug!("listening on {}", addr);
    tokio::spawn(loki_task);

    axum::Server::bind(&addr)
        .serve(app.into_make_service())
        .with_graceful_shutdown(shutdown_signal())
        .await?;

    tracing::info!("server down");

    opentelemetry::global::shutdown_tracer_provider();

    Ok(())
}

async fn shutdown_signal() {
    let ctrl_c = async {
        signal::ctrl_c()
            .await
            .expect("failed to install Ctrl+C handler");
    };

    #[cfg(unix)]
    let terminate = async {
        signal::unix::signal(signal::unix::SignalKind::terminate())
            .expect("failed to install signal handler")
            .recv()
            .await;
    };

    #[cfg(not(unix))]
    let terminate = std::future::pending::<()>();

    tokio::select! {
        _ = ctrl_c => {},
        _ = terminate => {},
    }

    tracing::info!("signal received, starting graceful shutdown");
}
